/**
 * Reads a diagram into the system from an access text file
 * @author Daniel Levy <dlevy at cs.toronto.edu>
 */


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;


public class ProcessDiagram {
	
	public ProcessDiagram()
	{
		
	}
	
	ArrayList <Node>tables; //Stores all the nodes. This includes there relations to other nodes
	
	
	/**
	 * Opens an ACCESS diagram text file, reads and processes the text file
	 * and stores the entities and there relationships 
	 *
	 * @param  fileName  The filename of the ACCESS diagram
	 * @return      A list of all graph edges and there relations
	 */
		public ArrayList <Node> process(String fileName) {			
			try
			{
				//read the file
				FileInputStream in = new FileInputStream(fileName);
				BufferedReader br = new BufferedReader(new InputStreamReader(in));
				String strLine;
				while((strLine = br.readLine())!= null)
				{
					// If its a table name then...
					if (strLine.compareToIgnoreCase("Relationships")  == 0 )
					{
						while (( (strLine = br.readLine())!= null) && (strLine.compareToIgnoreCase("Table Indexes") != 0))
						{
							String relate = isRelation(strLine);
							if (relate != null)
							{
								String tableNames[] = matchIt(relate);
								if (tableNames != null)
								{
									Node node1 = findNode(tableNames[0]);							
									Node node2 = findNode(tableNames[1]);
									node1.addNode(node2);
								}
							}							
						}
					}
				}											
			br.close();			
			}		
					
			catch(Exception e)
			{
				System.out.println(e);
			}
			return tables;
			
		}
		
		
		/**
		 * checks if a line of text represents a relation
		 * and stores the entities and there relationships 
		 *
		 * @param  strLine  A line of text from an ACCESS diagram text file
		 * @return     The line of text without whitespace, or NULL  if it is not a relation
		 */
		public String isRelation (String strLine)
		{
			if ((strLine.length()>5))
			{
				String temp = strLine.substring(0, 6);
				if ((temp.compareToIgnoreCase("Table:") != 0) 
						&& (temp.compareToIgnoreCase("c:\\Use") != 0))
				{
					String relate = strLine.trim();
					if (!hasWhiteSpace(relate))
					{
						return relate;
					}
				}
			}
			return null;
		}
		
		
		
		/**
		 * Takes in a string composed of two table names
		 * breaks it apart till it can match both the table names
		 * ie Table1Table2
		 *
		 * @param  str A string to test
		 * @return      An array with 2 elements, containing the 2 correct strings that represent tables
		 */
		public String[]  matchIt(String str)
		{
			String relate [] = new String[2];
			for (int i=0; i< str.length(); i++)
			{
				
				String test = str.substring(0, i);
				for (Node tabNode : tables)
				{
					String tabName = tabNode.className;
					
					if (tabName.compareToIgnoreCase(test)==0)
					{
						boolean matchedLastPart = false;
						String lastPart = str.substring(i, str.length());
						for (Node tabNode2 : tables)
						{
							if (lastPart.compareToIgnoreCase(tabNode2.className) == 0)
							{
								matchedLastPart = true;
								break;
							}
						}
						
						if (matchedLastPart)
						{
							relate[0] = test;
							relate[1] = str.substring(i, str.length());
							System.out.println(relate[0] + " ---- " + relate[1]);
							return relate;
						}
					}
				}
			}
			return null;
		}
		
		
		

		/**
		 * Reads an ACCESS diagram text file and 
		 *  locates all the table names contained, storing them 
		 *  in an ArrayList
		 *
		 * @param  fileName The filename of the ACCESS diagram
		 */
		public void listOfTables(String fileName) {
			tables = new ArrayList<Node>() ;
			String prevName = "";
			try
			{
				FileInputStream in = new FileInputStream(fileName);
				BufferedReader br = new BufferedReader(new InputStreamReader(in));
				String strLine;
				while((strLine = br.readLine())!= null)
				{
					if (strLine.length() < 6)
					{
						continue;
					}
					
					String temp = strLine.substring(0, 6);
					if (temp.compareToIgnoreCase("Table:") == 0)
					{
						//System.out.println(strLine.substring(0,40));
						String tableName = strLine.split(" ")[1];
						if (tableName.compareToIgnoreCase(prevName)!=0)
						{
							Node classNode = new Node(tableName);
							tables.add(classNode);
							prevName = tableName;
							//System.out.println(tableName);
						}
						
					}						
				}
				br.close();
			}
			
			catch(Exception e)
			{
				System.out.println(e);
			}
			
		}
		
		
		
		/**
		 * Looks for a node with the given node name in the table (ArrayList of nodes)
		 *
		 * @param  nodeName the name of the node we wish to locate
		 * @return      The located node
		 */
		public Node findNode (String nodeName)
		{
			for (Node tabNode : tables)
			{
				if (tabNode.className.compareToIgnoreCase(nodeName) == 0)
				{
					return tabNode;
				}
			}
			Node badNode = new Node ("badNode");
			return badNode;
		}
		
		
		
		/**
		 * Checks if a string contains whitespace or not
		 *
		 * @param  str A string to test
		 * @return      True if the string contains whitespace, false if not
		 */
		public boolean hasWhiteSpace(String str)
		{
			for (int i=0; i<str.length(); i++)
			{
				if (str.charAt(i) == ' ')
				{
					return true;
				}
			}
			return false;
		}

}
