/**
 * Contains the logic to find clusters in a graph. Implements a variation of 
 * the min-cut algorithm and uses wordNET similarity measures
 * @author Daniel Levy <dlevy at cs.toronto.edu>
 */



import java.util.ArrayList;

public class ProcessGraph {

	/**
	 * @param args
	 */
	
	ArrayList <Node>tables; //Stores all the nodes. This includes there relations to other nodes
	Util wordNetUtil;
	ArrayList <ArrayList<Node>> clusters;
	
	public ProcessGraph()
	{
		wordNetUtil = new Util();
		clusters = new ArrayList <ArrayList<Node>>();
	}	
	
	
	
	/**
	 * Receieves a node which includes its edges
	 * and finds ALL possible ways to combine the edges
	 * i.e if we have node 'A' with edges 1,2,3
	 * it will return 1,2,3,1 2, 1 3, 2 3, 1 2 3 
	 *  
	 * @param  node THe node for which we want to find the subsets
	 * @return      All possible subsets of the nodes relations
	 */
	public  ArrayList <ArrayList<Edge>> allNodeSubsets(Node node)
	{
		ArrayList <Edge>edgeNodes = node.edges;
		ArrayList <ArrayList<Edge>>edgeSubsets =  new ArrayList <ArrayList<Edge>>();
		for (Edge edge : edgeNodes)
		{
			ArrayList <ArrayList<Edge>>newEdgeSubsets =  new ArrayList <ArrayList<Edge>>();
			for (ArrayList<Edge> subsetEdge : edgeSubsets)
			{
				ArrayList<Edge> newEdge = copyList(subsetEdge);
				newEdge.add(edge);
				newEdgeSubsets.add(newEdge);
			}
			ArrayList <Edge> newSubset = new ArrayList <Edge>();
			newSubset.add(edge);
			edgeSubsets.add(newSubset);
			for (ArrayList<Edge> subsetEdge : newEdgeSubsets)
			{
				edgeSubsets.add(subsetEdge);
			}		
		}
		return edgeSubsets;
	}
	
	
	
	/**
	 * Recieves an arraylist and produces a new copy of it
	 *
	 * @param  list An ArrayList of edges that we wish to copy
	 * @return      An exact copy of the list
	 */
	public ArrayList <Edge> copyList (ArrayList<Edge> list)
	{		
		ArrayList<Edge> copiedList = new ArrayList<Edge>();
		for (Edge subsetEdge : list)
		{
			copiedList.add(subsetEdge);
		}
		return copiedList;
	}
	
	
	
	
	/**
	 * Recieves an arraylist and produces a new copy of it
	 *
	 * @param  list An ArrayList of edges that we wish to copy
	 * @return      An exact copy of the list
	 */
	public ArrayList <Node> copyNodeList (ArrayList<Node> list)
	{		
		ArrayList<Node> copiedList = new ArrayList<Node>();
		for (Node node : list)
		{
			copiedList.add(node);
		}
		return copiedList;
	}
	
	
	
	/**
	 * Tests if there is a way to traverse edges so that we reach node2 from node1
	 *
	 * @param  homeNode The originating node to test from
	 *  @param  destNode the destination node we want to see if we can reach
	 * @return      True if there is no way to reach destNode from homeNode (disconnected), false otherwise
	 */
	public boolean areDisconnected (Node homeNode, Node destNode)
	{
		if (homeNode.className.compareToIgnoreCase("Node5")==0)
		{
			boolean foundNode = false;
		}
		for (Node node : tables)
		{
			node.visited = false;
		}
		return areDisconnected2(homeNode, destNode);
	}
	
	
	
	/**
	 * Tests if there is a way to traverse edges so that we reach node2 from node1
	 *
	 * @param  homeNode The originating node to test from
	 *  @param  destNode the destination node we want to see if we can reach
	 * @return      True if there is no way to reach destNode from homeNode (disconnected), false otherwise
	 */
	public boolean areDisconnected2 (Node homeNode, Node destNode)
	{
		homeNode.visited = true;
		for (Edge tabNode : homeNode.edges)
		{
			homeNode.visited = true;
			Node compareNode = tabNode.getConnector(homeNode);
			if (compareNode == destNode)
			{
				return false;
			}
			else 
			{
				if (compareNode.visited == false)
				{
					boolean temp = areDisconnected2(compareNode, destNode);
					if (temp == false)
					{
						return false;
					}
				}
			}			
		}		
		return true;
	}
	
	
	
	/**
	 * Counts the number of nodes reachable from the input node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @return      The number of nodes reachable from the homenode
	 */
	//Returns the amount of nodes in this reachable from this cluster
	//Note - Does not count the nodes that form a discontinuity
	public int clusterSize(Node homeNode)
	{
		for (Node node : tables)
		{
			node.visited = false;
		}
		return clusterSize2(homeNode);
	}
	
	
	
	/**
	 * Counts the number of nodes reachable from the input node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @return      The number of nodes reachable from the homenode
	 */
	public int clusterSize2(Node homeNode)
	{
		int counter = 1;
		homeNode.visited = true;
		for (Edge tabNode : homeNode.edges)
		{
			Node compareNode = tabNode.getConnector(homeNode);
			if (compareNode.visited == false)
			{
				counter =  counter + clusterSize2(compareNode);
			}
		}		
		return counter;
	}
	
	

	/**
	 * Returns all the nodes reachable from this node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @return      A list of the reachable nodes
	 */
	public ArrayList <Node> getCluster(Node homeNode)
	{
		for (Node node : tables)
		{
			node.visited = false;
		}
		ArrayList <Node> cluster = new ArrayList <Node>();
		return getCluster2(homeNode, cluster);
	}
	
	
	
	/**
	 * Returns all the nodes reachable from this node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @return      A list of the reachable nodes
	 */
	public ArrayList <Node> getCluster2(Node homeNode, ArrayList <Node> cluster)
	{
		homeNode.visited = true;
		cluster.add(homeNode);	
		for (Edge tabNode : homeNode.edges)
		{
			Node compareNode = tabNode.getConnector(homeNode);
			if (compareNode.visited == false)
			{
				getCluster2(compareNode, cluster);
			}			
		}			
		return cluster;
	}	
	
	

	/**
	 * Identifies the sets of edges that represent min cuts in the diagram
	 * PSEUDOCODE:
	 * Iterate over all nodes
	 * 	Find all edge combinations for each node
	 * 	For each combination:
	 * 		Remove these edges from the graph
	 * 		Ensure all subset nodes are in the SAME cluster
	 * 		Test for a relation between the original node and the nodes in the subset
	 * 		Ensure that the cluster meets various conditions
	 * 		Restore the edges
	 *
	 * @return      A list of the mincuts
	 */
	public ArrayList <EdgeList> findMinCuts(ArrayList <Node> tables)
	{
		ArrayList <EdgeList> minCuts =  new ArrayList <EdgeList>(); //The list of mincuts
		
		//iterate over all nodes
		for (Node node : tables)
		{

			//get all the possible subset combinations of the edges
			ArrayList <ArrayList<Edge>> subsets =  allNodeSubsets(node);
			
			//iterate over the subsets for the specific node
			for (ArrayList <Edge> mincut : subsets)
			{
				//remove edges from the subset
				for (Edge removeEdge : mincut)
				{
					removeEdge.remove();
				}
				
				//initially claim this subset is a mincut
				boolean allDisconnect = true; 
				
				//iterate over the edges from THIS mincut
				//ensure that there is a disconnect between the basenode and ALL mincut edges
				for (int i = 0; i < mincut.size(); i++)
				{
					//get the node
					Node compareNode = mincut.get(i).getConnector(node);
					
					//test if there is a disconnection between the base node and the mincut  edges			
					if (!areDisconnected(node, compareNode))
					{
						allDisconnect = false;
						break;						
					}
					
					//ensure that all the mincut edges (that we know are disconnected from the basenode)
					// are all in the same cluster.
					// This is done so we will only find the mincut once. This mincut WILL be found in another subset if 
					// it does not meet these conditinos.
					if (i < mincut.size() -1)
					{
						Node nextNode = mincut.get(i+1).getConnector(node);
						
						if (areDisconnected2(compareNode, nextNode))
						{
							allDisconnect = false;
							break;
						}
					}
				}
				
				//This is a correct minCut
				// Now we need it to satisfy certain conditions.
				if (allDisconnect)
				{
					if (isMincut (minCuts, mincut, node))
					{					
						EdgeList list = new EdgeList(node, mincut);
						minCuts.add(list);
					}
				}
				
				//restore the edges
				for (Edge edge : mincut)
				{
					Node restoreNode = edge.getConnector(node);
					node.addNode(restoreNode);				
				}
			}			
		}
		return minCuts;
	}
	
	
	
	/**
	 * Verifies that the candidate mincut satisfies the conditions that make it a mincut
	 * Improvement is left for future works
	 *
	 * param  minCuts    The current list of approved minCuts
	 * param  mincut    The candidate mincut
	 * param  node    The node we are testing to see if it has a mincut   
	 * @return      A list of the mincuts
	 */
	public boolean isMincut (ArrayList <EdgeList> minCuts, ArrayList <Edge> mincut, Node node)
	{
		boolean isMinCut = false;
		
		//finds a node from the OPPOSING cluster (i.e the rest of the diagram)
		Node compareNode = mincut.get(0).getConnector(node); 
		
		//finds the smaller of the two clusters
		//(the one containing the base node vs a node from the opposing cluster)
		int smallCluster;
		int clustSize1 = clusterSize(node);
		int clustSize2 = clusterSize(compareNode);
		if (clustSize1 >= clustSize2)
		{
			smallCluster = clustSize2;
		}
		else
		{
			smallCluster = clustSize1;
		}
	
		//Future works  - better conditions
		//Checks that the smaller cluster contains at least 4 nodes
		if (smallCluster > 3)
		{
			isMinCut = true;	
		}			
		return isMinCut;
	}
	
	
	
	/**
	 * Process the mincuts and combine them in various ways 
	 * to find the clusters in the diagram.
	 * Will also assess a clusters viability through WordNET
	 * PSEUDOCODE:
	 * While we still have mincuts
	 * 	Iterate over each cut 'c'
	 * 		Temporarily remove the edges in c from the graph
	 * 		Check if this cluster leads to additional clusters
	 * 		If so, ignore it and restore edges
	 * 		Check if the cluster is a wordNET approved cluster
	 * 		If so, leave edges out and remove it from the list of cuts
	 * 		If not, mark the cluster for potential combination with other clusters
	 *
	 */
	public ArrayList <ArrayList<Node>> processCuts(ArrayList <Node> cluster, ArrayList <ArrayList<Node>> clusters)
	{
		
		ArrayList <EdgeList> cuts = findMinCuts(cluster);
		//Run until we have processed all cuts
		//Either ignoring them or 'made the cut'
		while (cuts.size()!=0)
		{
			//stores edgeLists we want to remove after exiting the loop	
			ArrayList<EdgeList> edgesToBeRemoved = new ArrayList<EdgeList>();
			
			//Get specific cuts
			for (EdgeList cut: cuts)
			{
				if (cut.edgeList.size() == 0)
				{
					continue;
					
				}
				//make the cut
				for (Edge edge : cut.edgeList)
				{
					edge.remove();
				}
				
				//Check which side to attempt to cluster
				Node checkNode = null;
				Node headNode = cut.head; //The headnode for the cut
				Node connector = cut.edgeList.get(0).getConnector(headNode); // A node in the 'opposing' cluster
				
				
				// check if we have additional cuts
				boolean hasMoreCutsHead = hasAdditionalCuts(headNode, cuts, cut);
				boolean hasMoreCutsConnector = hasAdditionalCuts(connector, cuts, cut);
				
				if ((!hasMoreCutsHead) && (hasMoreCutsConnector))
				{
					checkNode = headNode;
				}
				else if ((hasMoreCutsHead) && (!hasMoreCutsConnector))
				{
					checkNode = connector;
				}
				else if ((!hasMoreCutsHead) && (!hasMoreCutsConnector))
				{
					int sizeHead = clusterSize(headNode);
					int sizeConnector = clusterSize(connector);
					if (sizeHead <= sizeConnector)
					{
						checkNode = headNode;
					}
					else
					{
						checkNode = connector;
					}
				}
				
				boolean isCluster = false;
				//No additional cuts exiting from this cluster
				if (checkNode != null)
				{					
					edgesToBeRemoved.add(cut);//store the cut to be removed later
					removeEdgeFromMincuts(cuts, cut, edgesToBeRemoved);
					//Its the last cluster
					//Process both sides of it
					if (edgesToBeRemoved.size() - cuts.size() == 0)
					{			
						ArrayList <Node> connectorCluster = getCluster(connector);
						processCluster(connectorCluster);
					}		
					
					ArrayList <Node> nodeCluster = getCluster(checkNode);
					
					isCluster = processCluster(nodeCluster);				
				}
				
				//It still has more cuts
				//Restore the edges and process later
				if (!isCluster) 
				{
					if (clusters.size() == 1)
					{
						//ArrayList <Node> nodeCluster = getCluster(checkNode);
						clusters.add(cluster);
					}
					for (Edge edge : cut.edgeList)
					{
						edge.restore();
					}
				}				
			}
			
			if (cluster.size() >10)
			{
				if (cuts.size() == 2)
				{
					EdgeList cut2 = cuts.get(1);
					ArrayList <Node> connectorCluster = getCluster(cut2.head);
					clusters.add(connectorCluster);
					
				}
			}
			
			//remove the cuts that do not form clusters
			for (EdgeList removeCut : edgesToBeRemoved)
			{
				cuts.remove(removeCut);
			}			
			
		}
		return clusters;
	}
	
	
	
	public ArrayList <ArrayList <Node>> runSystem ()
	{
		clusters=  new ArrayList <ArrayList<Node>>();
		processCuts(tables, clusters);
		
		for (int i = 0; i <clusters.size(); i++)
		{
			ArrayList <Node> cluster = clusters.get(i);
			if (cluster.size() > 10)
			{
				for (Node node: cluster)
				{
					ArrayList<Node> nodeRemoveList = new ArrayList<Node>();
					for (int j = 0; j <node.edges.size(); j++)
					{
						Edge edge = node.edges.get(j);
						boolean inCluster = false;
						Node connectorNode = edge.getConnector(node);
						for (Node node2: cluster)
						{
							if (connectorNode == node2)
							{
								inCluster = true;
								break;
							}
						}
						if (!inCluster)
						{
							nodeRemoveList.add(connectorNode);							
						}
					}
					for (Node nodeRemove: nodeRemoveList)
					{
						node.removeEdge(nodeRemove);
					}
				}
				//ArrayList <ArrayList<Node>> newClusters = new ArrayList <ArrayList<Node>>();
				processCuts(cluster, clusters);
			}
		}
		return clusters;
	}
	
	/**
	 * Uses wordNet to evaluate if a cluster has sufficient inter-similiarity
	 *
	 * @param  cluster A group of nodes we wish to evaluate
	 * @return    Returns true if wordnet approves the cluster, flase otherwise
	 */
	public boolean processCluster(ArrayList <Node> cluster)
	{
		float nearnessMeasure = wordNetUtil.getAvgDistance(cluster);
		if (nearnessMeasure < 0.5) //future works  - automated threshold
		{
			clusters.add(cluster);
			return true;
		}
		else
		{
			return false;
		}
	}
	
	
	/**
	 * Removes an edge that has already been processed from all other mincuts
	 *
	 * @param  cuts The list of mincuts
	 * @param  cut The cut we are currently processing
	 * @param edgesToBeRemoved Store the EdgeLists so we can remove them later
	 */
	public void removeEdgeFromMincuts(ArrayList <EdgeList> cuts, EdgeList cut, ArrayList<EdgeList> edgesToBeRemoved)
	{
		for (int k = 0; k < cut.edgeList.size(); k ++)
		{					
			Edge comperator = cut.edgeList.get(k);
			
			//Remove this specific mincut from the other mincuts
			//ArrayList <EdgeList> tempEdgeList = new ArrayList <EdgeList>();
			for (EdgeList tempCut : cuts)
			{
				//skip this one, we will remove it later
				if (tempCut == cut)
				{
					continue;
				}
				if (removeEdge(comperator, tempCut.edgeList))
				{
					
					if (tempCut.edgeList.size() == 0)
					{
						edgesToBeRemoved.add(tempCut);
					}
				}
			}
		}
	}
	
	
	
	/**
	 * Attempts to remove an edge from an edgeList
	 *
	 * @param  compareEdge The  edge we are trying to remove 
	 * @param  edgeList the list we are trying to remove the edge from
	 * @return      True if we succesfuly removed the edge, false otherwise
	 */
	public boolean removeEdge(Edge compareEdge, ArrayList <Edge> edgeList)
	{
		for (Edge edge: edgeList)
		{
			if (edge.compare(compareEdge))
			{
				edgeList.remove(edge);
				return true;
			}
		}
		return false;
	}
	
	
	
	
	/**
	 * Checks if there are additional cuts reachable from  this node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @param  mincuts The list of mincuts
	 * @param  thisCut The cut we are currently processing (ignore it)
	 * @return      True if we can reach additional cuts from this node, false otherwise
	 */
	public boolean hasAdditionalCuts(Node homeNode, ArrayList <EdgeList> mincuts, EdgeList thisCut)
	{
		for (Node node : tables)
		{
			node.visited = false;
		}
		return hasAdditionalCuts2(homeNode, mincuts, thisCut);
	}
	
	
	
	/**
	 * Checks if there are additional cuts reachable from  this node
	 *  by traversing the nodes edges
	 *
	 * @param  homeNode The originating node to test from
	 * @param  mincuts The list of mincuts
	 * @param  thisCut The cut we are currently processing (ignore it)
	 * @return      True if we can reach additional cuts from this node, false otherwise
	 */
	public boolean hasAdditionalCuts2(Node homeNode, ArrayList <EdgeList> mincuts, EdgeList thisCut)
	{
		//REMEMBER TO IGNORE THISSSS MINCUT
		homeNode.visited = true;
		
		for (Edge nodeEdge: homeNode.edges)
		{
			if (isInMinCuts(nodeEdge, mincuts, thisCut))
			{
				return true;
			}
			else 
			{
				Node connector = nodeEdge.getConnector(homeNode);
				if (connector.visited == false)
				{
					boolean temp = hasAdditionalCuts2(connector, mincuts, thisCut);
					if(temp)
					{
						return true;
					}
				}
			}
		}

	
		return false;
	}
	
	
	/**
	 * Checks if a given edge is contained in the list of mincuts.
	 * Ignore 'thisCut' which is where the edge was originally found
	 *
	 * @param  edge The edge to test on
	 * @param  mincuts The list of mincuts
	 * @param  thisCut The cut we are currently processing (ignore it)
	 * @return      True if we can find this edge elsewhere in mincuts, false otherwise
	 */
	public boolean isInMinCuts(Edge edge, ArrayList <EdgeList> mincuts, EdgeList thisCut)
	{
		for (EdgeList edgeList : mincuts)
		{
			if (edgeList == thisCut)
			{
				continue;
			}
			
			for (Edge compareEdge : edgeList.edgeList)
			{
				if (compareEdge.compare(edge))
				{
					return true;
				}
			}
		}
		
		return false;
	}
	

	
}	







